# -*- coding: utf-8 -*-
#   This Source Code Form is subject to the terms of the Mozilla Public
#   License, v. 2.0. If a copy of the MPL was not distributed with this
#   file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""This module exposes key mixins to the larger module, letting this submodule deal with the details of the exposure."""
from .Action import ActionMixin
from .Context import ContextMixin
from .DeclarativeState import DeclarativeStateMixin
from .Plan import PlanMixin


# -*- coding: utf-8 -*-
#   This Source Code Form is subject to the terms of the Mozilla Public
#   License, v. 2.0. If a copy of the MPL was not distributed with this
#   file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Action base representing the smallest procedures for a Plan to carry out."""

from .Context import ContextMixin


class ActionMixin:
    """Abstract base representing the smallest procedures for a Plan to carry out.

    Actions are assembled by Plans and executed by Schedulers.
    Plans shoudl architect the overall procedure, but Actions MAY act as flowcontrol during plan execution.

    See "do".
    """

    def __init__(self, action_name: str, **kwargs):
        """Initialize an Action.

        Provide additional arguments in your constructor to configure your Actions.

        Args:
            action_name (str): A name uniquely identifying this action. Useful for logging.
            **kwargs: Provide additional arguments when implementing your Actions' constructors for configuration.

        """
        self.action_name = action_name
        super().__init__(**kwargs)

    # Must return another LIST of Actions or nothing
    # Must modify Context to reflect the actions taken
    def do(self, context: ContextMixin):
        """Do the action.

        Actions are assembled by Plans and executed by Schedulers. This is the entry hook into an Action.
        Actions MAY return a list of Actions to be executed following their own execution.
        Plans should architect the overall procedure, but this provides plan-runtime flow control.

        Args:
            context (ContextMixin): "do" references and may mutate Context, one step at a time.

        Raises:
            NotImplementedError: Must be implemented by child classes.

        Returns:
            list(ActionMixin): Optionally return a list of actions to be preformed after this one.

        """
        raise NotImplementedError

    def string(self):
        """Return a string representation of this Action for logging.

        Should state the intent of the action preformed.

        Returns:
            str: A string representation of this Action suitible for logging. Defaults to self.action_name.

        """
        return "{}".format(self.action_name)


if __name__ == "__main__":

    class __TestAction(ActionMixin):
        """Action for unit tests.

        Todo:
            Write better tests now that Contexts are finished.
        """

        def __init__(self, some: str, data: str, **kwargs):
            """Construct a Test Action with some member variables.

            Args:
                some (str): foo data
                data (str): bar data
            """
            self.some = some
            self.data = data
            super().__init__(**kwargs)

        def do(self, context):
            """Showcases actions' flowcontrol ability.

            Args:
                context: just a string for comparison.

            Returns:
                list(ActionMixin): Might return another TestAction.
            """
            # Do something with the context, in line with the name of the Action
            print(context == "TestVal")
            if context == "TestVal":
                return [__TestAction("Other", "Data", action_name="Recursive Action")]

    TA = __TestAction("Some", "Data", action_name="Test Context against TestVal")
    print(TA.do("TestVal")[0].do("TestVal")[0].do("Not TestVal"))



# -*- coding: utf-8 -*-
#   This Source Code Form is subject to the terms of the Mozilla Public
#   License, v. 2.0. If a copy of the MPL was not distributed with this
#   file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""This module exposes key configs to the larger module, letting this submodule deal with the details of the exposure."""

from .logging import ENABLE_DEVELOPMENT_LOGGING


# -*- coding: utf-8 -*-
#   This Source Code Form is subject to the terms of the Mozilla Public
#   License, v. 2.0. If a copy of the MPL was not distributed with this
#   file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Access to Logging facilities both for dev/debug and production purposes.

Library logging makes use of the python logging module's "extra" dict for
providing fields and their values for logging, and does not fill in "message".

This requires providing a format string that makes use of individual fields
for successfull logging.

This module provides read-only reference for the necessary Extra parameter
names required by any application-developer-supplied logging.formatter objects
used to extract logs from this library.

This module also provides a helper-function to enable basic logging for
library development.

Since this is a library, logging is disabled by default.
"""

import logging

##Defualt Configurations
pyReconcile_format = "%(asctime)s | %(name)s | %(levelname)s | %(message)s"
"""str: Default format string used for Library logging.

Used in :func:`ENABLE_DEVELOPMENT_LOGGING`
"""

pyReconcile_scheduler_format = (
    "%(asctime)s | %(name)s | %(levelname)s | %(run_id)s | %(type)s | %(description)s"
)
"""str: Format string used by the `pyReconcile.Scheduler` .

Used in :func:`ENABLE_DEVELOPMENT_LOGGING`
"""

## Since Scheduler relies on Extra arguments to log fields, Message is empty.
## Since Message is empty, propagating up to the default Logger results in useless log.
logging.getLogger("pyReconcile.Scheduler").propagate = False

## NullHandler prevents the library from logging until configured by the app-dev.
logging.getLogger("pyReconcile").addHandler(logging.NullHandler())


def ENABLE_DEVELOPMENT_LOGGING():
    """Enable logging for library development."""

    default_console_handler = logging.StreamHandler()
    default_console_handler.setLevel(logging.DEBUG)
    default_console_handler.setFormatter(logging.Formatter(pyReconcile_format))

    scheduler_console_handler = logging.StreamHandler()
    scheduler_console_handler.setLevel(logging.DEBUG)
    scheduler_console_handler.setFormatter(
        logging.Formatter(pyReconcile_scheduler_format)
    )

    logging.getLogger("pyReconcile").addHandler(default_console_handler)
    logging.getLogger("pyReconcile.Scheduler").addHandler(scheduler_console_handler)

    logging.getLogger("pyReconcile").setLevel(logging.DEBUG)
    logging.getLogger("pyReconcile.Scheduler").setLevel(logging.DEBUG)


# -*- coding: utf-8 -*-
#   This Source Code Form is subject to the terms of the Mozilla Public
#   License, v. 2.0. If a copy of the MPL was not distributed with this
#   file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Abstract base that carries out orchestrating one iteration of a reconciliation procedure."""
from .Context import ContextMixin
from .Action import ActionMixin


class PlanMixin:
    """Abstract base that carries out orchestrating one iteration of a reconciliation procedure.

    Plans create lists of Actions to be executed sequentially by Schedulers.
    Schedulers will continue to create and execute lists of Actions until an empty list is returned.
    """

    def __init__(self, plan_name: str, **kwargs):
        """Initialize a Plan.

        Optionally provide additional kwargs in your constructor to inform your action plan creation.

        Args:
            plan_name (str): A name uniquely identifying this plan. Useful for logging.
            **kwargs: Provide additional arguments when implementing your Plans' constructors for configuration.

        """
        self.name = plan_name
        super().__init__(**kwargs)

    # Must return an Iterable of Actions
    def create(self, context: ContextMixin):
        """Create a list of actions based on a given context.

        Schedulers use Plans to create the overall sequence of actions necessary to reconcile a Context's two states.
        This MUST return an None or an empty list when states are reconciled, or the Scheduler will go infininte.

        ContextMixin member functions, like get_state_diff, should be used here to inform plan creation.
        See ContexMixin.

        Args:
            context (ContextMixin): The context used to inform plan creation.

        Raises:
            NotImplementedError: Must be implemented to act as a Plan for a Scheduler.

        Returns:
            list(ActionMixin): The list of actions to be carried out in sequence or None when context states are reconciled.
        """
        raise NotImplementedError


if __name__ == "__main__":

    class __TestPlan(PlanMixin):
        def __init__(self, some, data, **kwargs):
            self.some = some
            self.data = data
            super().__init__(**kwargs)

        def create(self, context):
            # Do something with the context, to create a list of actions that need to be taken
            if context[0]:
                return ["step 1"]
            else:
                return ["step 0", "step 1", "step 2"]

    TP = __TestPlan("Some", "Data", plan_name="My Plan")
    print(TP.create([False]))



# -*- coding: utf-8 -*-
#   This Source Code Form is subject to the terms of the Mozilla Public
#   License, v. 2.0. If a copy of the MPL was not distributed with this
#   file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""The core representation of state for reconciliation.

Makes use of DeepDiff to provide a solid representation of the differences
that need to be reconciled between two instances of DeclarativeStates.

Some DeepDiff kwargs can be passed through the constructor for
the instance's diffing against another instance. This is not symmetric.
"""

from deepdiff import DeepDiff


class DeclarativeStateMixin(object):
    """The core representation of state for reconciliation."""

    def __init__(
        self,
        declarative_state_data,
        declarative_state_ignore_order=True,
        declarative_state_report_repetition=False,
        **kwargs
    ):
        """Construct a DeclarativeState around any data supported by DeepDiff.

        DeepDiff's kwargs provided here are used in this instance's diff_declarative_state call.

        Args:
            declarative_state_data: Data representing state. Must be supported by DeepDiff.
            declarative_state_ignore_order (bool, optional): See DeepDiff's docs for details. Defaults to True.
            declarative_state_report_repetition (bool, optional): See DeepDiff's docs for details. Defaults to False.
        """
        self.state = declarative_state_data
        self.__ignore_order = declarative_state_ignore_order
        self.__report_repetition = declarative_state_report_repetition
        self.__verbose_level = 2
        super().__init__(**kwargs)

    @property
    def state(self):
        """Access the internal State data. Useful for fetching data for use by Actions."""
        return self.__state

    @state.setter
    def state(self, val):
        self.__state = val

    def diff_declarative_state(self, other) -> DeepDiff:
        """Run DeepDiff against another DeclarativeState.

        Used when comparing the Desired State against the Current State in a Context.
        Uses this instance's DeepDiff kwargs from this instance's constructor.

        Args:
            other (DeclarativeStateMixin): The other DeclarativeState to diff against.

        Returns:
            DeepDiff: The DeepDiff between this and other.
        """
        return DeepDiff(
            self.state,
            other.state,
            ignore_order=self.__ignore_order,
            report_repetition=self.__report_repetition,
            verbose_level=self.__verbose_level,
        )


if __name__ == "__main__":
    print("Testing DeclarativeState Mixin")
    t1 = {1: 1, 2: 2, 3: 3, 4: {"a": "hello", "b": [1, 2, 3]}, 5: 5}
    t2 = {1: 1, 2: 2, 3: 3, 4: {"a": "hello", "b": [1, 3, 2, 3]}}
    print(t1)
    print(t2)

    class __Test(DeclarativeStateMixin):
        def __init__(self, name, *args, **kwargs):
            super().__init__(**kwargs)
            self.name = name

    ignore_order = True
    report_repetition = True
    a1 = __Test(
        "Name is a1", declarative_state_data=t1
    )  # , declarative_state_ignore_order=ignore_order, declarative_state_report_repetition=report_repetition)
    a2 = __Test(
        "Name is a2", declarative_state_data=t2
    )  # , declarative_state_ignore_order=ignore_order, declarative_state_report_repetition=report_repetition)
    print(a1.diff_declarative_state(a2))
    d = a1.diff_declarative_state(a2)
    b1 = __Test(
        "Name is b1",
        declarative_state_data=[1, 2, 3],
        declarative_state_ignore_order=False,
        declarative_state_report_repetition=True,
    )
    b2 = __Test(
        "Name is b2",
        declarative_state_data=[1, 3, 3, 4],
        declarative_state_ignore_order=False,
        declarative_state_report_repetition=True,
    )
    b3 = __Test(
        "Name is b3",
        declarative_state_data=[1, 3],
        declarative_state_ignore_order=False,
        declarative_state_report_repetition=True,
    )
    b4 = __Test(
        "Name is b4",
        declarative_state_data=[1, 2, 3],
        declarative_state_ignore_order=False,
        declarative_state_report_repetition=True,
    )
    print(b1.diff_declarative_state(b2))
    print(b1.state)
    b1.state[1] = 3
    b1.state.append(4)
    print(b1.state)
    print(b1.diff_declarative_state(b2))
    print(b4.diff_declarative_state(b3))



# -*- coding: utf-8 -*-
#   This Source Code Form is subject to the terms of the Mozilla Public
#   License, v. 2.0. If a copy of the MPL was not distributed with this
#   file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Collects two DeclarativeStates to represent the current and desired states to reconcile."""

from deepdiff import DeepDiff

from .DeclarativeState import DeclarativeStateMixin


class ContextMixin:
    """Collects two DeclarativeStates to represent the current and desired states to reconcile."""

    # Desired State and Current State MUST implement Declarative State
    def __init__(
        self,
        desired_state: DeclarativeStateMixin,
        current_state: DeclarativeStateMixin,
        **kwargs
    ):
        """Collects the desired_state and current_state for later use in diffing.

        Args:
            desired_state (DeclarativeStateMixin): The state a Plan should attempt to mutate the current_state into.
            current_state (DeclarativeStateMixin): The current state that needs to be reconciled to the desired one.
        """
        assert isinstance(desired_state, DeclarativeStateMixin)
        assert isinstance(current_state, DeclarativeStateMixin)
        self.desired_state = desired_state
        self.current_state = current_state
        super().__init__(**kwargs)

    def update_current_state(self, **kwargs):
        """Optional. Provide a way to update the current_state, like querying a REST API."""
        pass

    def update_desired_state(self, **kwargs):
        """Optional. Provide a way to update the desired_state, like querying a REST API."""
        pass

    def get_state_diff(self) -> DeepDiff:
        """Diffs the current state against the desired state.

        Returns:
            DeepDiff: The DeepDiff of the Current State against the Desired State.
        """
        return self.current_state.diff_declarative_state(self.desired_state)


# -*- coding: utf-8 -*-
#   This Source Code Form is subject to the terms of the Mozilla Public
#   License, v. 2.0. If a copy of the MPL was not distributed with this
#   file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Schedulers execute Plans on a given Context.

Logging:
    See config.logging.
    
    Message: No message is logged, fields are provided through the logging module's Extra dict.
    "Extra" keys:
        run_id: a UUID to correlate Plan execution runs.
        type: "Plan" or "Action". Plans can create multiple action lists per execution run. Actions done are logged.
        description: Message context. Plan Name or Action.string().

"""

import uuid
import logging

from typing import List

from .mixins.Plan import PlanMixin
from .mixins.Context import ContextMixin
from .mixins.Action import ActionMixin

# from .config.logging import logging_preprocessors

logger = logging.getLogger(__name__)


class Scheduler:
    """Schedulers execute Plans on a given Context.

    Plans are executed until they return an empty list or None from Plan.create.
    """

    def __init__(self, name: str, plans: List[PlanMixin] = []):
        """Currently doesn't do anything special, might collect Plans in the future. Not sure.

        Args:
            name (str): NYI
            plans (List[PlanMixin], optional): NYI, schedulers might collect Plans. Not sure. Defaults to [].
        """
        self.name = name
        self.plans = plans
        self.run_id = None

    # Execute a given plan with context
    def execute(self, plan: PlanMixin, context: ContextMixin):
        """Execute a given plan with context.

        Plans are executed until they return an empty list or None from Plan.create.

        Args:
            plan (PlanMixin): The plan to execute. See plan.create(context).
            context (ContextMixin): The context to provide to plan.create for building our list of Actions to do.

        Logging:
            Logs each plan iteration, using Plan.name for "description".
        """

        # step_count = 0
        self.run_id = str(uuid.uuid1())
        # We create our list of steps for our iteration
        # Since Actions update Context, running a plan many times MUST produce a context that requires no actions to be performed
        # WARN This is Dangerous, this means you MUST VALIDATE that your ACTIONS produce CONTEXTS that ARE RECONCILED, but that's kinda the point, no?
        steps = plan.create(context)
        while steps:
            # step_count += 1
            logger.info(
                "",
                extra={"run_id": self.run_id, "type": "plan", "description": plan.name},
            )
            self.react(context, steps)
            steps = plan.create(context)

    # Recursively execute Actions (which may return further actions, requiring the recursion)
    def react(self, context: ContextMixin, steps: List[ActionMixin]):
        """Recurively do Actions (which may return further actiosn, requiring the recursion).

        Args:
            context (ContextMixin): The Context to pass to Action.do
            steps (List[ActionMixin]): The Actions being done.

        Logging:
            Logs each step, using Action.string() for "description".
        """
        for step in steps:
            logger.info(
                "",
                extra={
                    "run_id": self.run_id,
                    "type": "action",
                    "description": step.string(),
                },
            )
            result = step.do(context)
            if result:
                self.react(context, result)



# -*- coding: utf-8 -*-
#   This Source Code Form is subject to the terms of the Mozilla Public
#   License, v. 2.0. If a copy of the MPL was not distributed with this
#   file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""This module exposes key features of submodules at the namespace of the library.

The library is small, and names exclusive enough to make this interface clean.
"""
from .config import *
from .mixins import *
from .Scheduler import Scheduler


# -*- coding: utf-8 -*-
#   This Source Code Form is subject to the terms of the Mozilla Public
#   License, v. 2.0. If a copy of the MPL was not distributed with this
#   file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Package Details."""

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyReconcile",  # Replace with your own username
    version="0.0.2",
    author="Charles Unruh",
    author_email="charles.unruh.dev@gmail.com",
    description="A reconciliation loop system.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/cunruh3760/pyreconcile",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries",
    ],
    python_requires=">=3.6.8",
    install_requires=["deepdiff==5.2.2"],
)